/* ======================================================================
     "bx1S2HRsol.cc"
       - Solve time-independent Schroedinger equation for a symmetric
          two-dimensional hindered rotor by basis function expansion
          (CP2: for cos(theta) power series 2-D potential and Legendre
          polynomial basis)
             Copyright (c) 2004-2012  by A. Miyoshi, Univ. Tokyo
                         created: June  25, 2006
                          edited: July  19, 2006
                     last edited: Mar.  12, 2012
 ------------------------------------------------------------------------
 usage
  $ bx1S2HRsol basefilename
  - input is read from 'basefilename'.inp

 ====================================================================== */

#include "bx1S2HRsol.h"

/* ===== main function ================================================ */

int main(int argc, char *argv[]) {
  string basfln;
  int outCtrl = OUTnone, precanal = true;
  int fullJmax, redJmax, maxRel;
  double rtolEigen;
  tempList templ;
  basisAssocLegendre *basp;
  eigenVal2Dsrot redEgn;

  showRevision();
  if (argc != 2) { errms(ERRusage); return 1; }
  basfln = argv[1];
  basp = new basisAssocLegendre;
  if (read_input(basfln, *basp, templ, outCtrl, precanal, rtolEigen)
   == ERRE) { return 1; }
  fullJmax = basp->maxJdeg; redJmax = fullJmax * 0.7;

  if (precanal) {
     // --- solve with reduced max J
    basp->maxJdeg = redJmax;
    if (basp->solveOnly() == ERRE) { errms(NUMerr); return 1; }
    redEgn = basp->egnVal;
     // --- solve with full max J and output vectors & functions
    basp->maxJdeg = fullJmax;
    if (basp->solveWrite(basfln, outCtrl) == ERRE)
     { errms(NUMerr); return 1; }
    if (detMaxRel(basp->egnVal, redEgn, maxRel, rtolEigen) == ERRE)
     { return 1; }
  } else {
     // --- no precision analysis
    basp->maxJdeg = fullJmax;
    if (basp->solveWrite(basfln, outCtrl) == ERRE)
     { errms(NUMerr); return 1; }
    maxRel = -1;
  }
  if (basp->writeEnergies(basfln, maxRel) == ERRE) { return 1; }
  if (basp->writePartFuncs(basfln, maxRel, templ) == ERRE) { return 1; }
  delete basp;
  return 0;
}

/* ----- read input --------------------------------------------------- */

int read_input(string &basfln, basisAssocLegendre &basis, tempList &tpls,
 int &outCtrl, int &precanal, double &rtolEigen) {
  int symn, iss, noc, ioc, blkend, ppord;
  double ppcoef;
  string inpfln = basfln + ".inp", key;
  vector<string> vs, vsv;
  ifstream inf;
  int maxJ, outJmax;
  double rotC;
  nucSpinStat nsw;
  powerSeries pe;
  int maxJ_read = false, rotC_read = false, outJmax_read = false;
  int rtolEigen_read = false;

  inf.open(inpfln.c_str());
  if (!inf) { return errms(FAILopen, inpfln); }
  pe.clear(); tpls.clear();
  while (inpTokenKeyValues(inf, key, vs) == NORM) {
    if (key == "maxJ") {
      maxJ_read = true; maxJ = atoi(vs[0].c_str());
    } else if (key == "rotConst") {
      rotC_read = true; rotC = atof(vs[0].c_str());
    } else if (key == "nuSpinSt") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVnuss); }
      if (vs.size() <= symn) { return errms(INVnuss); }
      nsw.clear();
      for (iss = 0; iss < symn; iss++)
       { nsw.push_back(atoi(vs[iss + 1].c_str())); }
    } else if (key == "symNumbr") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVsymn); }
      nsw.clear();
      for (iss = 0; iss < symn; iss++) { nsw.push_back(1); }
    } else if (key == "output") {
      noc = vs.size();
      for (ioc = 0; ioc < noc; ioc++) {
        if (vs[ioc] == "vector") { outCtrl |= OUTvec; }
        else if (vs[ioc] == "function") { outCtrl |= OUTfunc; }
        else if (vs[ioc] == "all") { outCtrl |= OUTall; }
      }
    } else if (key == "outJmax") {
      outJmax_read = true; outJmax = atoi(vs[0].c_str());
    } else if (key == "precAnal") {
      if (vs[0] == "yes") { precanal = true; }
      else if (vs[0] == "no") { precanal = false; }
    } else if (key == "rtolEigen") {
      rtolEigen_read = true; rtolEigen = atof(vs[0].c_str());
    } else if (key == "tempRange") {
      if (tpls.addRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempRecipRange") {
      if (tpls.addRecipRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempList") {
      if (tpls.addListVS(vs) == ERRE) { return ERRE; }
    } else if (key == "potPars{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else if (vsv.size() < 2) { return errms(INVpotp); }
        else {
          ppord = atoi(vsv[0].c_str());
          ppcoef = atof(vsv[1].c_str());
          pe.insert(make_pair(ppord, ppcoef));
        }
      }
    } else if (key != "#") { errms(UNKNkey, key); }
  }
  inf.close();
  if (!maxJ_read) { return errms(NOmaxJ); }
  if ((maxJ <= 0) || (maxJ >= basis.MXNBAS)) { return errms(INVmaxJ); }
  if (!outJmax_read) { outJmax = maxJ * 0.3; }
  if (outJmax < 0) { outJmax = 0; }
  else if (outJmax > maxJ) { outJmax = maxJ; }
  if (!rtolEigen_read) { rtolEigen = 0.01; }
  if (rtolEigen < 1.e-8) { rtolEigen = 1.e-8; }
  else if (rtolEigen > 0.1) { rtolEigen = 0.1; }
  if (!rotC_read) { return errms(NOrotc); }
  if (rotC <= 0.) { return errms(INVrotc); }
  if (pe.size() == 0) { return errms(NOVpotp); }
  if (nsw.size() == 0) { nsw.push_back(1); }
  if (tpls.empty()) { tpls.setDefault(); }
  basis.init(maxJ, outJmax, rotC, nsw, pe);
  return NORM;
}

/* ----- determine maximum reliable number of eigen states ------------ */

int detMaxRel(eigenVal2Dsrot &fullEgn, eigenVal2Dsrot &redEgn,
 int &maxRel, double rtolEigen) {
  int nFull = fullEgn.qnTable.size(), nRed = redEgn.qnTable.size();
  int nRelRed = nRed, ie;
  double vf, vr, fracRel;
  for (ie = 0; ie < nRed; ie++) {
    vf = fullEgn.getEgnv(ie); vr = redEgn.getEgnv(ie);
    if (fabs(vf) >= VAL0THR) {
      if (fabs((vr - vf) / vf) > rtolEigen) { nRelRed = ie; break; }
    }
  }
  fracRel = (double)nRelRed / (double)nRed;
  maxRel = nFull * fracRel - 1;
  cout << "estimated eigenvalue error is <= " << rtolEigen
   << " at number = " << maxRel << ".\n";
  if ((maxRel < 3) || (fracRel < (sqrt(rtolEigen) * 0.5) ))
   { return errms(FATALconv); }
  else if ((maxRel < 10) || (fracRel < sqrt(rtolEigen)))
   { errms(BADconv); }
  return NORM;
}

/* ----- error message functions -------------------------------------- */

int errms(ecode ec, string ms) {
  string s;
  switch(ec) {
    case ERRusage: s = "usage: bx1S2HRsol basefilename"; break;
    case FAILopen: s = "- Failed to open [" + ms + "]."; break;
    case READerr: s = "- File read error."; break;
    case UNKNkey: s = ".. Ignored unknown key [" + ms + "]."; break;
    case INVnuss: s = "- Invalid nuSpinSt input."; break;
    case INVsymn: s = "- Invalid symNumbr input."; break;
    case INVpotp: s = "- Invalid potPars input."; break;
    case NOmaxJ: s = "- No maxJ input found."; break;
    case INVmaxJ: s = "- Invalid maxJ input."; break;
    case NOrotc: s = "- No rotConst input found."; break;
    case INVrotc: s = "- Invalid rotConst input."; break;
    case NOVpotp: s = "- No valid potPars input found."; break;
    case NUMerr: s = "- Numerical error."; break;
    case FATALconv: s = "- Extremely bad convergence. Fatal."; break;
    case BADconv: s = "warning: Bad convergence."; break;
  }
  cout << s << endl;
  return ERRE;
}

