/* ======================================================================
     "bx1VIBsol.cc"
       - Solve one-dimensional time-independent Schroedinger equation
          by basis function expansion (PS: for power series potential
          and Hermite-Gaussian basis)
             Copyright (c) 2004-2012  by A. Miyoshi, Univ. Tokyo
                         created: Sept.  8, 2005
                          edited: Sept. 20, 2005
                          edited: Oct.  14, 2005
                     last edited: Mar.  12, 2012
 ------------------------------------------------------------------------
 - Rev. 0.11 : added nuSpinSt/symNumbr input
 ------------------------------------------------------------------------
 usage
  $ bx1VIBsol basefilename
  - input is read from 'basefilename'.inp

 ====================================================================== */

#include "bx1VIBsol.h"

/* ===== main function ================================================ */

int main(int argc, char *argv[]) {
  string basfln;
  int outcnt = OUTnone, fnmBas, smlBas, limQn;
  powerSeries potE;
  tempList templ;
  basisHermiteGauss *basp;   // - large object cannot be an autovariable

  if (argc != 2) { errms(ERRusage); return 1; }
  basfln = argv[1];
  basp = new basisHermiteGauss;
  double egvS[basp->MXNBAS];
  if (read_input(basfln, potE, *basp, templ, outcnt) == ERRE)
   { return 1; }
  fnmBas = basp->numBasis();
  smlBas = (fnmBas - 1) / 2 + 1;
  if (small_run(potE, *basp, smlBas, egvS) == ERRE)
   { errms(NUMerr); return 1; }

  basp->setNumBasis(fnmBas);
  if (basp->solve(potE) == ERRE) { errms(NUMerr); return 1; }

  limQn = max_reliable_qn(*basp, smlBas, egvS);
  cout << "estimated eigenvalue error is 1% at quntum number = "
   << limQn << ".\n";
  if ((limQn < 5) || ((double)limQn / (fnmBas - 1) < 0.05))
   { errms(FATALconv); return 1; }
  else if ((limQn < 20) || ((double)limQn / (fnmBas - 1) < 0.10))
   { errms(BADconv); }

  if (write_eigen_values(basfln, *basp, limQn, fnmBas - 1) == ERRE)
   { return 1; }
  if (write_part_funcs(basfln, *basp, limQn, templ) == ERRE) { return 1; }
  if (outcnt & OUTvec) {
    if (write_eigen_vectors(basfln, *basp, limQn) == ERRE) { return 1; }
  }
  if (outcnt & OUTfunc) {
    if (write_eigen_funcs(basfln, *basp, limQn) == ERRE) { return 1; }
  }
  delete basp;
  return 0;
}

/* ----- read input --------------------------------------------------- */

int read_input(string &basfln, powerSeries &potE,
 basisHermiteGauss &basis, tempList &tpls, int &outcnt) {
  ifstream inf;
  string inpfln = basfln + ".inp", key;
  vector<string> vs, vsv;
  int nbas_read = false, rotc_read = false, hof_read = false,
   potp_read = false;
  int nbas, npp = 0, blkend, noc, ioc, symn = 0, iss;
  double rotc, hof, alp;
  powerTerm powt;
  nucSpinStat nspin;

  showRevision();
  inf.open(inpfln.c_str());
  if (!inf) { return errms(FAILopen, inpfln); }
  potE.clear(); tpls.clear();
  while (inpTokenKeyValues(inf, key, vs) == NORM) {
    if (key == "numBasis") {
      nbas_read = true; nbas = atoi(vs[0].c_str());
    } else if ((key == "rotConst") || (key == "hbar2/2m")) {
      rotc_read = true; rotc = atof(vs[0].c_str());
    } else if (key == "hoFreq") {
      hof_read = true; hof = atof(vs[0].c_str());
    } else if (key == "nuSpinSt") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVnuss); }
      if (vs.size() <= symn) { return errms(INVnuss); }
      nspin.clear();
      for (iss = 0; iss < symn; iss++)
       { nspin.push_back(atoi(vs[iss + 1].c_str())); }
    } else if (key == "symNumbr") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVsymn); }
      nspin.clear();
      for (iss = 0; iss < symn; iss++) { nspin.push_back(1); }
    } else if (key == "output") {
      noc = vs.size();
      for (ioc = 0; ioc < noc; ioc++) {
        if (vs[ioc] == "vector") { outcnt |= OUTvec; }
        else if (vs[ioc] == "function") { outcnt |= OUTfunc; }
        else if (vs[ioc] == "all") { outcnt |= OUTall; }
      }
    } else if (key == "tempRange") {
      if (tpls.addRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempRecipRange") {
      if (tpls.addRecipRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempList") {
      if (tpls.addListVS(vs) == ERRE) { return ERRE; }
    } else if (key == "potPars{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else if (vsv.size() < 2) { return errms(INVpotp); }
        else {
          powt.order = atoi(vsv[0].c_str());
          powt.coef = atof(vsv[1].c_str());
          potE.push_back(powt); npp++;
        }
      }
      if (npp > 0) { potp_read = true; potE.setMaxOrder(); }
    } else if (key != "#") { errms(UNKNkey, key); }
  }
  inf.close();

  if (!nbas_read) { return errms(NOnbas); }
  if (nbas <= 0) { return errms(INVnbas); }
  if (!rotc_read) { return errms(NOrotc); }
  if (rotc <= 0.) { return errms(INVrotc); }
  if (!hof_read) { return errms(NOhof); }
  if (hof <= 0.) { return errms(INVhof); }
  if (!potp_read) { return errms(NOVpotp); }
  alp = sqrt(2 * rotc / hof);
  if (symn == 0) { nspin.clear(); nspin.push_back(1); }
  if (tpls.empty()) { tpls.setDefault(); }
  basis.setParam(nbas, potE.maxOrder, rotc, alp, nspin);
  return NORM;
}

/* ----- run with reduced nBasis -------------------------------------- */

int small_run(powerSeries &potE, basisHermiteGauss &basis, int snbas,
 double *egvs) {
  int iq;
  double *eg;
  basis.setNumBasis(snbas);
  if (basis.solve(potE) == ERRE) { return ERRE; }
  eg = basis.getEigenValues();
  for (iq = 0; iq < snbas; iq++) { egvs[iq] = eg[iq]; }
  return NORM;
}

/* ----- evaluate the maximum reliable quantum number ----------------- */

int max_reliable_qn(basisHermiteGauss &basis, int snbas, double *egvs) {
  int iq, ilimq = -1, limqfull, nfullbas = basis.numBasis();
  double *eg = basis.getEigenValues();
  for (iq = 0; iq < snbas; iq++) {
    if (fabs((egvs[iq] - eg[iq]) / eg[iq]) > 0.01) { ilimq = iq; break; }
  }
  if (ilimq == -1) { ilimq = snbas; }
  ilimq--; if (ilimq < 0) { ilimq = 0; }
  limqfull = (int)(((double)ilimq / (snbas - 1)) * (nfullbas - 1));
  return limqfull;
}

/* ----- write eigen values ------------------------------------------- */

int write_eigen_values(string &basfln, basisHermiteGauss &basis,
 int mxQn, int mxOutQn) {
  int iq, symn = basis.nSpinW.size(), iss, sumSw = 0;
  double *egval = basis.getEigenValues();
  ofstream ouf;
  string outfln = basfln + "_eigen_values.csv";
  outFormat *ofmt;

  for (iss = 0; iss < symn; iss++) { sumSw += basis.nSpinW[iss]; }
  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmt = new outFormat(&ouf);
  ouf << "number,eigenValue,w\n";
  ouf.precision(8);
  for (iq = 0; iq <= mxOutQn; iq++) {
    iss = iq % symn;
    ouf << iq << "," << egval[iq] << ",(" << basis.nSpinW[iss] << "/"
     << sumSw << ")";
    if (iq == mxQn) { ouf << ",(maximum reliable qntNbr)"; }
    ouf << endl;
  }
  delete ofmt;
  ouf.close();
  return NORM;
}

/* ----- write eigen vectors ------------------------------------------ */

int write_eigen_vectors(string &basfln, basisHermiteGauss &basis,
 int mxQn) {
  int iq, nBasis = basis.numBasis(), ib;
  int nPages = mxQn / 250 + 1, jp, qbegin, qlast;
  double *egval = basis.getEigenValues();
  double *egvcc;
  ofstream ouf;
  string outfln;
  outFormat *ofmt;

  for (jp = 0; jp < nPages; jp++) {
    qbegin = jp * 250; qlast = jp * 250 + 249;
    if (qlast > mxQn) { qlast = mxQn; }
    outfln = basfln + "_eigen_vectors" + format(jp, 2, '0') + ".csv";
    ouf.open(outfln.c_str());
    if (!ouf) { return errms(FAILopen, outfln); }
    ofmt = new outFormat(&ouf);
    ouf << "eigen_values"; ouf.precision(8);
    for (iq = qbegin; iq <= qlast; iq++) { ouf << "," << egval[iq]; }
    ouf << "\neigen_vectors" ;
    for (iq = qbegin; iq <= qlast; iq++) { ouf << ",vec" << iq; }
    ouf << endl; ouf.precision(4);
    for (ib = 0; ib < nBasis; ib++) {
      ouf << "c" << ib;
      egvcc = basis.getEgnVecCoefs(ib);
      for (iq = qbegin; iq <= qlast; iq++)
        { ouf << "," << egvcc[iq]; }
      ouf << endl;
    }
    delete ofmt;
    ouf.close();
  }
  return NORM;
}

/* ----- write eigen functions ---------------------------------------- */

int write_eigen_funcs(string &basfln, basisHermiteGauss &basis,
 int mxQn) {
  int iq;
  int nPages = mxQn / 250 + 1, jp, qbegin, qlast;
  double ymax = sqrt(1.8 * (basis.numBasis() - .5)) + 5.;
  double ystep = .02, ymin = -ymax, alp = basis.HGalpha(), x, y;
  double *egnf;
  vector<ofstream *> ouf(nPages);
  ofstream *ouf1;
  string outfln;
  vector<outFormat *> ofmt(nPages);
  outFormat *ofmt1;

  ouf.clear(); ofmt.clear();
  for (jp = 0; jp < nPages; jp++) {
    qbegin = jp * 250; qlast = jp * 250 + 249;
    if (qlast > mxQn) { qlast = mxQn; }
    outfln = basfln + "_eigen_funcs" + format(jp, 2, '0') + ".csv";
    ouf1 = new ofstream(outfln.c_str());
    if (!(*ouf1)) { return errms(FAILopen, outfln); }
    ouf.push_back(ouf1);
    ofmt1 = new outFormat(ouf[jp]);
    ofmt.push_back(ofmt1);
    *ouf[jp] << "x";
    for (iq = qbegin; iq <= qlast; iq++) { *ouf[jp] << ",f" << iq; }
    *ouf[jp] << endl;
  }
  for (y = ymin; y <= ymax; y += ystep) {
    x = y * alp;
    egnf = basis.getEigenFunc(y, mxQn);
    for (jp = 0; jp < nPages; jp++) {
      qbegin = jp * 250; qlast = jp * 250 + 249;
      if (qlast > mxQn) { qlast = mxQn; }
      ouf[jp]->precision(6); *ouf[jp] << x; ouf[jp]->precision(4);
      for (iq = qbegin; iq <= qlast; iq++)
        { *ouf[jp] << "," << egnf[iq]; }
      *ouf[jp] << endl;
    }
  }
  for (jp = 0; jp < nPages; jp++) {
    delete ofmt[jp]; ouf[jp]->close(); delete ouf[jp];
  }
  return NORM;
}

/* ----- calculate and write partition functions ---------------------- */

int write_part_funcs(string &basfln, basisHermiteGauss &basis, int mxQn,
 tempList &tpls) {
  double *eg = basis.getEigenValues();
  double Tmax, TmaxInL, T, errPf, kT, Qpf;
  int itl, ntl = tpls.size();
  int iq, symn = basis.nSpinW.size(), iss, sumSw = 0;
  ofstream ouf;
  string outfln = basfln + "_part_funcs.csv";
  outFormat *ofmf;

  for (iss = 0; iss < symn; iss++) { sumSw += basis.nSpinW[iss]; }
  Tmax = eg[mxQn] * KELWAVN / log(100.);
  TmaxInL = tpls.maxLowerEq(Tmax);
  if (TmaxInL < 0.) { return errms(LRGQerr); }
  errPf = exp(-eg[mxQn] * KELWAVN / TmaxInL);
  cout << "estimated partition function error is " << errPf << " at "
   << TmaxInL << " K.\n";
  if (TmaxInL < 2000.) { errms(LMTQrng); }

  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmf = new outFormat(&ouf);
  ouf.precision(8);
  ouf << "10000/T,T,Q\n";
  for (itl = 0; itl < ntl; itl++) {
    T = tpls[itl];
    if (T <= Tmax) {
      kT = T / KELWAVN; Qpf = 0.;
      for (iq = mxQn; iq >= 0; iq--) {
        iss = iq % symn;
        Qpf += (double)(basis.nSpinW[iss]) * exp(-eg[iq] / kT);
      }
      Qpf /= (double)sumSw;
      ouf << (10000. / T) << "," << T << "," << Qpf << endl;
    }
  }
  delete ofmf;
  ouf.close();
  return NORM;
}

/* ----- error message functions -------------------------------------- */

int errms(ecode ec, string ms) {
  string s;
  switch(ec) {
    case ERRusage: s = "usage: bx1VIBsol basefilename"; break;
    case FAILopen: s = "- Failed to open [" + ms + "]."; break;
    case READerr: s = "- File read error."; break;
    case UNKNkey: s = ".. Ignored unknown key [" + ms + "]."; break;
    case NOnbas: s = "- No numBasis input found."; break;
    case INVnbas: s = "- Invalid numBasis input."; break;
    case NOrotc: s = "- No rotConst (or hbar2/2m) input found."; break;
    case INVrotc: s = "- Invalid rotConst (or hbar2/2m) input."; break;
    case NOhof: s = "- No hoFreq input found."; break;
    case INVhof: s = "- Invalid hoFreq input."; break;
    case INVsymn: s = "- Invalid symNumbr input."; break;
    case INVnuss: s = "- Invalid nuSpinSt input."; break;
    case INVpotp: s = "- Invalid potPars input."; break;
    case NOVpotp: s = "- No valid potPars input found."; break;
    case NUMerr: s = "- Numerical error."; break;
    case BADconv: s = "warning: Bad convergence."; break;
    case FATALconv: s = "- Extremely bad convergence. Fatal."; break;
    case LRGQerr: s = "- No partfunc calc can be performed."; break;
    case LMTQrng: s = "warning: Limited T-range for Q-calc."; break;
  }
  cout << s << endl;
  return ERRE;
}

