/* ======================================================================
     "libbasIOstr.cc"
       - basic I/O and string library
             Copyright (c) 2002-2012  by A. Miyoshi, Univ. Tokyo
      created from "gpoplibf.cc": Sept.  8, 2005
                          edited: Sept. 18, 2005
                     last edited: Mar.  12, 2012
 ====================================================================== */

#include "libbasIOstr.h"

/* ======================================================================
     I/O functions
 ====================================================================== */

/* ----- output formatting class -------------------------------------- */

outFormat::outFormat() { ostrm = NULL; }
outFormat::outFormat(ostream *os)
 { ostrm = os; flagSave = ostrm->flags(); prec = ostrm->precision(); }
outFormat::~outFormat() { restore(); }
void outFormat::restore()
 { ostrm->precision(prec); ostrm->flags(flagSave); }

/* ----- show revision number ----------------------------------------- */

void showRevision() { cout << BEX1DREV << endl; }

/* ----- get a line from a stream ------------------------------------- */

int getLine(istream &is, string &line) {
  char c;
  line = "";
  if (is.get(c).eof()) { return ERRE; }
  do {
    if (c == '\n') { break; }
    if (c != '\r') { line += c; }
  } while (!is.get(c).eof());
  return NORM;
}

/* ----- get a nonblank line after trimming --------------------------- */

int getLineNonBlank(istream &is, string &line) {
  do {
    if (getLine(is, line) == ERRE) { return ERRE; }
    line = trim(line);
  } while (line.empty());
  return NORM;
}

/* ----- get a valid line after comment-removal and trimming ---------- */

int getLineValid(istream &is, string &line, string cmtsep) {
  string::size_type pcmt;
  do {
    if (getLine(is, line) == ERRE) { return ERRE; }
    if ((pcmt = line.find_first_of(cmtsep)) != string::npos)
      { line = line.substr(0, pcmt); }
    line = trim(line);
  } while (line.empty());
  return NORM;
}

/* ----- tokenize input (treat !/# as comment/title head char) -------- */

int inpToken(istream &is, vector<string> &vs) {
  string line;
  if (getLineValid(is, line, "!") == ERRE) { return ERRE; }
  if (line[0] == '#')
   { vs.clear(); vs.push_back("#"); vs.push_back(ltrim(line.substr(1))); }
  else { vs = splitTokens(line, " \t,"); }
  return NORM;
}

/* ----- tokenize input to key and value strings ---------------------- */

int inpTokenKeyValues(istream &is, string &key, vector<string> &vs) {
  if (inpToken(is, vs) == ERRE) { return ERRE; }
  key = vs[0]; vs.erase(vs.begin());
  return NORM;
}

/* ======================================================================
     string manipulation functions
 ====================================================================== */

/* ----- convert all characters to upper case ------------------------- */

string strAllToUpper(string &s) {
  string rs = s;
  string::size_type ls = rs.length(), p;
  for (p = 0; p < ls; p++) { rs[p] = toupper(rs[p]); }
  return rs;
}

/* ----- convert all characters to lower case ------------------------- */

string strAllToLower(string &s) {
  string rs = s;
  string::size_type ls = rs.length(), p;
  for (p = 0; p < ls; p++) { rs[p] = tolower(rs[p]); }
  return rs;
}

/* ----- trim white space chars --------------------------------------- */

string ltrim(string s) {
  string::size_type p1st;
  if ((p1st = s.find_first_not_of(" \t\r\n")) != string::npos)
    { return s.substr(p1st); }
  return "";
}

string rtrim(string s) {
  string::size_type plast;
  if ((plast = s.find_last_not_of(" \t\r\n")) != string::npos)
    { return s.substr(0, plast + 1); }
  return "";
}

string trim(string s) { return rtrim(ltrim(s)); }

/* ----- split string into tokens ------------------------------------- */

vector<string> splitTokens(string &s, const string &sep) {
  string::size_type p1st, pNxSep, pStart = 0, ls = s.length();
  string tkn;
  vector<string> vs;
  vs.clear();
  while ((p1st = s.find_first_not_of(sep, pStart)) != string::npos) {
    if ((pNxSep = s.find_first_of(sep, p1st)) == string::npos) {
      tkn = s.substr(p1st); vs.push_back(tkn); break;
    } else {
      tkn = s.substr(p1st, pNxSep - p1st); vs.push_back(tkn);
      pStart = pNxSep;
    }
  }
  return vs;
}

/* ----- prepare formatted string of an numerical value --------------- */

string format(double v) {
  ostringstream os;
  os << v;
  return os.str();
}

string format(int v) {
  ostringstream os;
  os << v;
  return os.str();
}

string format(int v, int len, char filc) {
  ostringstream os;
  os << setw(len) << setfill(filc) << v;
  return os.str();
}

/* ----- get base part from a filename -------------------------------- */

string getBasePart(string flname) {
  string basname;
  string::size_type p;
  if ((p = flname.rfind('.')) != string::npos) {
    basname = flname.substr(0, p);
  } else {
    basname = flname;
  }
  return basname;
}

/* ----- get extension part from a filename --------------------------- */

string getExtPart(string flname) {
  string extname;
  string::size_type p;
  if ((p = flname.rfind('.')) != string::npos) {
    extname = flname.substr(p + 1);
  } else {
    extname.erase();
  }
  return extname;
}

