/* ======================================================================
     "libbx1com.cc"
       - Linear least squares library
             Copyright (c) 2004-2005  by A. Miyoshi, Univ. Tokyo
                         created: Oct.  14, 2005
   (part of the codes imported from gpoplibt.cc/.h)
                     last edited: Oct.  28, 2005   Rev. 0.10
 ====================================================================== */

#include "libbx1com.h"

/* ======================================================================
     temperature list
 ====================================================================== */

tempList::tempList() { setDefault(); }            // --- constructor ---

void tempList::setDefault()                        // --- setDefault ---
  { clear(); addRange(300., 1500., 100.); }

void tempList::sortElimDup() {   // --- sort & eliminate duplication ---
  int id, jd, nelm = size();
  double d, dlast;
  vector<double> vd = *this;
  if (nelm < 2) { return; }
  for (id = 0; id < (nelm - 1); id++) {
    for (jd = (id + 1); jd < nelm; jd++) {
      if (vd[jd] < vd[id]) { d = vd[id]; vd[id] = vd[jd]; vd[jd] = d; }
    }
  }
  clear(); push_back(vd[0]); dlast = vd[0];
  for (id = 1; id < nelm; id++) {
    if ((vd[id] - dlast) > 1.e-10) { push_back(vd[id]); dlast = vd[id]; }
  }
}
                                                  // --- add a range ---
int tempList::addRange(double st, double et, double dt) {
  double t;
  if (dt < 1.e-6) { return errms(INVRstep); }
  if (st < 1.e-6) { return errms(INVRstrt); }
  if (et < st) { return errms(INVRinvt); }
  if (((et - st) / dt) > 1.e4) { return errms(INVRtmny); }
  for (t = st; t < (et + dt * .1); t += dt) { push_back(t); }
  sortElimDup();
  return NORM;
}
                                     // --- add a range from strings ---
int tempList::addRangeVS(vector<string> &vs, int offset) {
  // - set offset = 1 when vs[0] contains key "tempRange" etc.
  double st, et, dt;
  if ((offset < 0) || (offset > 100)) { return errms(BADoffst); }
  if (vs.size() < (offset + 3)) { return errms(INVRnpar); }
  st = atof(vs[offset].c_str()); et = atof(vs[offset + 1].c_str());
  dt = atof(vs[offset + 2].c_str());
  return addRange(st, et, dt);
}
                                   // --- add a range in reciprocals ---
int tempList::addRecipRange(double numr, double stt, double lst,
 double stp) {
  double rt;
  if (numr < 1.e-12) { return errms(INVInumr); }
  if (stp < 1.e-12) { return errms(INVIstep); }
  if (stt < 1.e-12) { return errms(INVIstrt); }
  if (lst < stt) { return errms(INVIinvt); }
  if (((lst - stt) / stp) > 1.e4) { return errms(INVItmny); }
  for (rt = stt; rt < (lst + stp * .1); rt += stp)
   { push_back(numr / rt); }
  sortElimDup();
  return NORM;
}
                      // --- add a range in reciprocals from strings ---
int tempList::addRecipRangeVS(vector<string> &vs, int offset) {
  // - set offset = 1 when vs[0] contains key "tempRecipRange" etc.
  double numr, stt, lst, stp;
  if ((offset < 0) || (offset > 100)) { return errms(BADoffst); }
  if (vs.size() < (offset + 4)) { return errms(INVInpar); }
  numr = atof(vs[offset].c_str()); stt = atof(vs[offset + 1].c_str());
  lst = atof(vs[offset + 2].c_str()); stp = atof(vs[offset + 3].c_str());
  return addRecipRange(numr, stt, lst, stp);
}
                                                   // --- add a list ---
int tempList::addList(vector<double> &vd) {
  int id, nd = vd.size();
  for (id = 0; id < nd; id++) {
    if (vd[id] < 1.e-6) { return errms(INVLtemp); }
    push_back(vd[id]);
  }
  sortElimDup();
  return NORM;
}
                                      // --- add a list from strings ---
int tempList::addListVS(vector<string> &vs, int offset) {
  // - set offset = 1 when vs[0] contains key "tempList" etc.
  int is, ns = vs.size();
  vector<double> vd;
  if ((offset < 0) || (offset > 100)) { return errms(BADoffst); }
  vd.clear();
  for (is = offset; is < ns; is++) { vd.push_back(atof(vs[is].c_str())); }
  if (!vd.empty()) { return addList(vd); }
  return NORM;
}

double tempList::maxTemp() {              // --- maximum Temperature ---
  int itl, ntl = size();
  double maxT = 0.;
  for (itl = 0; itl < ntl; itl++) {
    if ((*this)[itl] > maxT) { maxT = (*this)[itl]; }
  }
  return maxT;
}
                          // --- max temp lower than or equal to 't' ---
double tempList::maxLowerEq(double thT) {
  int itl, ntl = size();
  double maxT = -1.;
  for (itl = 0; itl < ntl; itl++) {
    if (((*this)[itl] > maxT) && ((*this)[itl] <= thT))
     { maxT = (*this)[itl]; }
  }
  return maxT;
}
                                                // --- error message ---
int tempList::errms(errcode erc, string sv) {
  string msg = "tempList: ";
  switch(erc) {
    case INVRstep: msg += "invalid range (step < 1E-6)."; break;
    case INVRstrt: msg += "invalid range (start < 1E-6)."; break;
    case INVRinvt: msg += "invalid range (end < start)."; break;
    case INVRtmny: msg += "invalid range (n > 1E+4)."; break;
    case INVRnpar: msg += "invalid range (< 3 parameters)."; break;
    case INVInumr: msg += "invalid range reciplocal (numerator < 1E-12)."; break;
    case INVIstep: msg += "invalid range reciplocal (step < 1E-12)."; break;
    case INVIstrt: msg += "invalid range reciplocal (start < 1E-12)."; break;
    case INVIinvt: msg += "invalid range reciplocal (end < start)."; break;
    case INVItmny: msg += "invalid range reciplocal (n > 10000)."; break;
    case INVInpar: msg += "invalid range reciplocal (< 3 parameters)."; break;
    case BADoffst: msg += "bad offset (< 0 or >100)."; break;
    case INVLtemp: msg += "invalid list (T < 1E-6)."; break;
  }
  cout << msg << endl;
  return ERRE;
}

