/* ======================================================================
     "libnumLls_dsysv.h"
       - header file for "libnumLls_dsysv.cc"
             Copyright (c) 2005-2007  by A. Miyoshi, Univ. Tokyo
 ====================================================================== */

#ifndef _LIBNUMLLS_DSYSV_H_
#define _LIBNUMLLS_DSYSV_H_

#include "libbasIOstr.h"
#include "lapack_dsysv/wcpp_dsysv.h"

/* ===== basic Linear Least Squares Class ============================= */

class basicLinLsSq {
 public:
  enum { MXNPAR = 169,      // maximum number of fitting parameters
         MXLWRK = 10816 };  // max length of working array = MXNPAR * 64

  /* MXNPAR > 181 causeed unexpected problem
     probably because MXNPAR ** 2 > 32767    */

 protected:
  int nPars;             // number of parameters
  int nMeas;             // number of 'measurement' points
  enum erract { INIT, DADD, SOLV };
  enum errcode { NOT_init, LRG_nPars, SML_nPars, SMLN_x,
                 SMLN_meas, meas_SMLTp,
                 POSVpar, POSVfail, POTRIpar, POTRIfail };
 private:
  vector<vector<double> > sumXX;
  vector<double> sumYX;
  double sumYY;
  double a[MXNPAR][MXNPAR], b[1][MXNPAR];
  int ipiv[MXNPAR];
  double work[MXLWRK];
 public:
  basicLinLsSq();
  basicLinLsSq(int npar);
  void clear();
  int init(int npar);
  int addDatumW(double y, vector<double> &x, double w);
  int addDatum(double y, vector<double> &x);
  int putDataW(int n, vector<double> &y, vector<vector<double> > &x,
   vector<double> &w);
  int putData(int n, vector<double> &y, vector<vector<double> > &x);
  int solve_blls(vector<double> &mlp, vector<double> &sdp,
   vector<vector<double> > &pcc);
 protected:
  void errmsg(erract era, errcode erc);
};

/* ===== Multiple Dependent-Variable Linear Least Squares
          and Related Classes                             ============= */

// --- function unit ---------------------------------------------------

class funcUnit {
 public:
  string fname;         // function name
  int order;            // function order
  double coef;          // function coefficient
 public:
  funcUnit();
  funcUnit(string nam, int ord, double cf);
  void clear();
  int init(string nam, int ord, double cf);
  int isProperlySet();
  string termStr();
  double val(double x);
  funcUnit differnFunc();
};

// --- list of funcUnit's = a sigle variable function ------------------

class listFnUnits : public vector<funcUnit> {
 public:
  int add(string nam, int ord, double cf = 1.);
  int addNoOrd(string nam, double cf = 1.);
  int addFrom(vector<string> &vs);
  listFnUnits differnFunc();
};

// --- power or Fourier series function (used as output) ---------------

class powFourUnit {
 public:
  int order;
  double coef, param;
};

class powFourSer : public vector<powFourUnit> {};

// --- data point set (used as input) ----------------------------------

class pointXYW {
 public:
  double x, y, w;
};

class dataPntsXYW : public vector<pointXYW> {};

// --- multiple dependent-variable Linear Least Squares ----------------

class multiVarLLS : protected basicLinLsSq {
 protected:
  enum loc_errcod { INVinitPAR, NOTinitOBJ, INVfuncNBR, INVfuncLST,
                    INSnumCOEF, NOTsetGFNC, INVdegDIFF, NOTsetFNCS };
 private:
  vector<listFnUnits> optFuncs;
  vector<double> optCoef;
  vector<listFnUnits> fixFuncs;
  vector<double> fixCoef;
 public:
  multiVarLLS();
  multiVarLLS(int ndv);
  void clear();
  int init(int ndv);
  int numDpndVars();
  int set1OptFnc(int fnbr, listFnUnits &lfu);
  int add1OptFncTrm(int fnbr, string nam, int ord, double cf = 1.);
  int add1OptFncTrmNoOrd(int fnbr, string nam, double cf = 1.);
  int add1OptFncTrmFrom(int fnbr, vector<string> &vs);
  int set1FixFncCoef(int fnbr, listFnUnits &lfu, vector<double> &ffc);
  int add1FixFncTrm(int fnbr, string nam, int ord, double cf = 1.);
  int add1FixFncTrmNoOrd(int fnbr, string nam, double cf = 1.);
  int add1FixFncTrmFrom(int fnbr, vector<string> &vs);
  void addFixCoef(double fc);
  int setDerivFunc(int fnbr, int gnbr, int deg);
  int numTermsOpt();
  int numTermsFix();
  string funcStr(int fnbr);
  vector<string> funcsStrAll();
  vector<double> getFixCoef();
  int isFuncSet();
  int addPoint(int fnbr, double x, double y, double w = 1.);
  int setPoints(int fnbr, dataPntsXYW &dps);
  int solve(vector<double> &mlp, vector<double> &sdp,
   vector<vector<double> > &pcc);
  int wasSolved();
  int isPowerSer(int fnbr, powFourSer &ps);
  int isFourierSer(int fnbr, powFourSer &fs);
 protected:
  int loc_err(loc_errcod lerc);
};

#endif
